from typing import Any, Dict, List, Optional
from pydantic import Field
from langchain_core.retrievers import BaseRetriever
from langchain_core.documents import Document
from langchain_core.callbacks import CallbackManagerForRetrieverRun
from langchain_igpt._igpt_search_api_wrapper import IgptSearchAPIWrapper
import warnings

class IgptSearchRetriever(BaseRetriever):
    """
    iGPT Search API retriever.

    Provides retrieval over the user's connected personal datasources,
    including emails, messages, files, attachments, and metadata.


    Setup:
        Install `langchain-igpt` and set environment variables `IGPT_API_KEY`, `IGPT_API_USER`.

        ```bash
        pip install -U langchain-igpt
        export IGPT_API_KEY="your-api-key"
        export IGPT_API_USER="your-api-user"
        ```

    Key init args:
        date_from:(string, optional)
            Start date filter ('YYYY-MM-DD')
        date_to (string, optional)
            End date filter ('YYYY-MM-DD').
        max_results: (number, optional)
            Limit number of results (e.g., 50)

    Instantiate:
        ```python
        from langchain_igpt import IgptSearchRetriever
        retriever = IgptSearchRetriever(max_results=50)
        ```

    Example:
        ```python
        docs = retriever.invoke("contract renewal date")
        ```

    """

    api_wrapper: IgptSearchAPIWrapper = Field(
        default_factory=IgptSearchAPIWrapper
    )

    # Search params (max_results, date filters, etc.)
    params: Dict[str, Any] = Field(default_factory=dict)

    def __init__(self, user: Optional[str] = None, api_key: Optional[str] = None, base_url: Optional[str] = None, max_retries: int = 3, backoff_factor: int = 2, backoff_base: int = 100, **kwargs: Any) -> None:
        data = {}
        if api_key or base_url or user:
            wrapper_kwargs = {}
            if api_key:
                wrapper_kwargs["api_key"] = api_key
            if base_url:
                wrapper_kwargs["base_url"] = base_url.rstrip("/")
            if user:
                wrapper_kwargs["user"] = user
                
            wrapper_kwargs["max_retries"] = max(1, max_retries) if isinstance(max_retries, int) else 3
            wrapper_kwargs["backoff_factor"] = max(2, backoff_factor) if isinstance(backoff_factor, int) else 2

            backoff_base_ms = max(100, backoff_base) if isinstance(backoff_base, int) else 100
            wrapper_kwargs["backoff_base"] = backoff_base_ms / 1000.0 # 0.1 seconds
            data["api_wrapper"] = IgptSearchAPIWrapper(**wrapper_kwargs)

        data["params"] = dict(kwargs or {})

        super().__init__(**data)

    def _get_relevant_documents(
        self,
        query: str,
        *, # This forces all following arguments to be keyword-only
        run_manager=CallbackManagerForRetrieverRun, # This is LangChain’s callback tracing system
        user: Optional[str] = None,
        **kwargs: Any,
    ) -> List[Document]:
        params = dict(self.params)
        params.update(kwargs)
        params["query"] = query

        try:
            response = self.api_wrapper.search(user=user, **params)
            if response.get("context") and not response["context"].get("datasources"):
                warnings.warn(
                    "No datasources found. Please authorize and connect a new datasource at "
                    "https://igpt.ai/hub/playground/connectors-authorize/"
                )

            return self._parse_response(response)
        except Exception as e:
            return {"error": e}

    # Async retrieval
    async def _aget_relevant_documents(
        self,
        query: str,
        *,
        run_manager: CallbackManagerForRetrieverRun,
        user: Optional[str] = None,
        **kwargs: Any,
    ) -> List[Document]:
        params = dict(self.params)
        params.update(kwargs)
        params["query"] = query

        try:
            response = await self.api_wrapper.search_async(user=user, **params)
            if response.get("context") and not response["context"].get("datasources"):
                warnings.warn(
                    "No datasources found. Please authorize and connect a new datasource at "
                    "https://igpt.ai/hub/playground/connectors-authorize/"
                )
            return self._parse_response(response)
        except Exception as e:
            return {"error": e}


    # Documents mapping
    def _parse_response(self, response: Dict[str, Any]) -> List[Document]:
        """
        Convert iGPT search response into LangChain Documents.

        Preserves:
        - id
        - timestamp
        - type
        - from or name
        - subject
        - source (normalized)
        """

        results = response.get("results", [])
        documents: List[Document] = []

        for item in results:
            content = item.get("content") or ""
            metadata = {k: v for k, v in item.items() if k != "content"} 
            if not metadata.get("source"):
                metadata["source"] = self._build_source(item)

            documents.append(
                Document(
                    page_content=content,
                    metadata=metadata,
                )
            )

        return documents
    
    def _build_source(self, item: Dict[str, Any]) -> str:
        """
        Build a stable, human-readable source identifier.
        """
        item_type = str(item.get("type") or "unknown").strip()
        item_id = str(item.get("id") or "").strip()
        return f"{item_type}:{item_id}" if item_id else item_type


from typing import Any, Dict, Optional, AsyncGenerator, Generator, Union
import time
from pydantic import BaseModel, ConfigDict, SecretStr, model_validator
from langchain_core.utils import get_from_dict_or_env
import requests
import aiohttp
import asyncio
import json
import warnings
from importlib.metadata import version

IGPT_BASE_URL: str = "https://api.igpt.ai/v1"

class IgptAskAPIWrapper(BaseModel):
    """Wrapper for Igpt Ask API."""

    api_key: SecretStr
    base_url: Optional[str] = None
    user: Optional[str] = None
    max_retries: int = 3
    backoff_factor: int = 2
    backoff_base: float = 0.1 # 100 milliseconds
    _version = version("langchain-igpt")

    model_config = ConfigDict(
        extra="forbid",
    )

    @model_validator(mode="before")
    def validate_environment(cls, values: Dict) -> Any:
        api_key = get_from_dict_or_env(
            values, "api_key", "IGPT_API_KEY"
        )
        values["api_key"] = api_key
        user = get_from_dict_or_env(
            values, "user", "IGPT_API_USER"
        )
        values["user"] = user
        return values

    def ask(self, user: Optional[str] = None, **kwargs: Any) -> Union[Dict[str, Any], Generator[dict, None, None]]:
        payload = {}
        payload["user"] = user or self.user or None
        if kwargs:
            payload.update(kwargs)

        base_url = self.base_url or IGPT_BASE_URL

        url = f"{base_url}/recall/ask/"
        return self._stream_request(url, payload) if payload.get("stream") else self._post(url, payload)

    def _post(self, url: str, payload: dict):
        headers = {
            "Authorization": f"Bearer {self.api_key.get_secret_value()}",
            "Content-Type": "application/json",
            "X-Client": f"langchain-igpt/{self._version}"
        }
        delay = self.backoff_base
        for attempt in range(self.max_retries):
            try:
                resp = requests.post(url, json=payload, headers=headers, timeout=(60, 600)) # connect_timeout 60 seconds, read_timeout 10 minutes
                if resp.status_code == 200:
                    try:
                        return resp.json()
                    except Exception:
                        return {"error": "invalid_json_response"}
                elif resp.status_code == 429 or 500 <= resp.status_code < 600: # Server-side error (429 - Too Many Requests) - retry
                    pass
                else:
                    return {"error": "client_error"} # Client-side error - don't retry
            except requests.exceptions.RequestException: # Network error or timeout - retry
                pass

            if attempt < self.max_retries - 1: # Sleep before next retry, except after the last attempt
                time.sleep(delay)
                delay *= self.backoff_factor  # Exponential backoff

        return {"error": "network_error"}
    
    def _stream_request(self, url: str, payload: dict) -> Generator[dict, None, None]:
        headers = {
            "Authorization": f"Bearer {self.api_key.get_secret_value()}",
            "Content-Type": "application/json",
            "X-Client": f"langchain-igpt/{self._version}"
        }
        delay = self.backoff_base
        for attempt in range(self.max_retries):
            try:
                with requests.post(url, json=payload, stream=True, headers=headers, timeout=(60, 600)) as resp: 
                    if resp.status_code == 200:
                        for chunk in resp.iter_lines():
                            if chunk:
                                chunk_str = chunk.decode()
                                if chunk_str.startswith("data:"):
                                    json_str = chunk_str[len("data:"):].strip()
                                else:
                                    json_str = chunk_str

                                if not json_str:
                                    continue
                                
                                try:
                                    data = json.loads(json_str)
                                    if data.get("context") and not data["context"].get("datasources"):
                                        warnings.warn(
                                            "No datasources found. Please authorize and connect a new datasource at " 
                                            "https://igpt.ai/hub/playground/connectors-authorize/"
                                        )
                                    yield data
                                except json.JSONDecodeError:
                                    continue
                        return
                    elif resp.status_code == 429 or 500 <= resp.status_code < 600:
                        pass
                    else:
                        yield {"error": "client_error"}
                        return
            except requests.exceptions.RequestException:
                pass

            if attempt < self.max_retries - 1:
                time.sleep(delay)
                delay *= self.backoff_factor
            else:  
                yield {"error": "network_error"}

    # Async
    async def ask_async(self, user: Optional[str] = None, **kwargs: Any) -> Union[Dict[str, Any], AsyncGenerator[dict, None]]:
        payload = {}
        payload["user"] = user or self.user or None
        if kwargs:
            payload.update(kwargs)

        base_url = self.base_url or IGPT_BASE_URL

        url = f"{base_url}/recall/ask/"
        return self._stream_request_async(url, payload) if payload.get("stream") else await self._post_async(url, payload)

    async def _post_async(self, url: str, payload: dict):
        headers = {
            "Authorization": f"Bearer {self.api_key.get_secret_value()}",
            "Content-Type": "application/json",
            "X-Client": f"langchain-igpt/{self._version}"
        }

        delay = self.backoff_base
        timeout = aiohttp.ClientTimeout(
            total=None,
            connect=60,      # connect timeout (sec)
            sock_read=600,   # read timeout (sec) = 10 min
        )
        async with aiohttp.ClientSession(headers=headers, timeout=timeout) as session:
            for attempt in range(self.max_retries):
                try:
                    async with session.post(url, json=payload) as resp:
                        if resp.status == 200:
                            try:
                                return await resp.json()
                            except Exception:
                                return {"error": "invalid_json_response"}
                        elif resp.status == 429 or 500 <= resp.status < 600: # Server-side error (429 - Too Many Requests) - retry
                            pass
                        else:
                            return {"error": "client_error"} # Client-side error - don't retry
                except (aiohttp.ClientError, asyncio.TimeoutError):
                    pass

                if attempt < self.max_retries - 1: # Sleep before next retry, except after the last attempt
                    await asyncio.sleep(delay)
                    delay *= self.backoff_factor # Exponential backoff

        return {"error": "network_error"}
    
    async def _stream_request_async(self, url: str, payload: dict) -> AsyncGenerator[dict, None]:
        headers = {
            "Authorization": f"Bearer {self.api_key.get_secret_value()}",
            "Content-Type": "application/json",
            "X-Client": f"langchain-igpt/{self._version}"
        }
        delay = self.backoff_base
        timeout = aiohttp.ClientTimeout(
            total=None,
            connect=60,    # connect timeout (sec)
            sock_read=600, # read timeout (sec): max idle time between chunks
        )
        async with aiohttp.ClientSession(headers=headers, timeout=timeout) as session:
            for attempt in range(self.max_retries):
                try:
                    async with session.post(url, json=payload) as resp:
                        if resp.status == 200:
                            async for chunk in resp.content:
                                if chunk:
                                    chunk_str = chunk.decode()
                                    if chunk_str.startswith("data:"):
                                        json_str = chunk_str[len("data:"):].strip()
                                    else:
                                        json_str = chunk_str
                                        
                                    if not json_str:
                                        continue
                                    
                                    try:
                                        data = json.loads(json_str)
                                        if data.get("context") and not data["context"].get("datasources"):
                                            warnings.warn(
                                                "No datasources found. Please authorize and connect a new datasource at " 
                                                "https://igpt.ai/hub/playground/connectors-authorize/"
                                            )
                                        yield data
                                    except json.JSONDecodeError:
                                        continue
                            return
                        elif resp.status == 429 or 500 <= resp.status < 600:
                            pass
                        else:
                            yield {"error": "client_error"}
                            return
                except (aiohttp.ClientError, asyncio.TimeoutError):
                    pass

                if attempt < self.max_retries - 1:
                    await asyncio.sleep(delay)
                    delay *= self.backoff_factor
                
            yield {"error": "network_error"}



from typing import Any, Dict, Optional, Type, Union, Literal, AsyncGenerator, Generator
from langchain_core.tools import BaseTool, ToolException
from pydantic import BaseModel, ConfigDict, Field
from langchain_igpt._igpt_ask_api_wrapper import IgptAskAPIWrapper
import warnings

class IgptAskInput(BaseModel):
    model_config = ConfigDict(extra="allow")

    input: str = Field(
        description=("The question to ask")
    )

class IgptAsk(BaseTool):
    """

    Setup:
        Install `langchain-igpt` and set environment variables `IGPT_API_KEY`, `IGPT_API_USER`.

        ```bash
        pip install -U langchain-igpt
        export IGPT_API_KEY="your-api-key"
        export IGPT_API_USER="your-api-user"
        ```

    Instantiate:

        ```python
        from langchain_igpt import IgptAsk

        tool = IgptAsk(quality="cef-1-normal")
        ```

    Invoke directly with args:

        ```python
        tool.invoke({"input": "Summarize my last meeting in 5 bullet points.", "output_format":"text"})
        ```

        ```json
       {
            "id": "...",
            "output": "...",
            "context": {
                "quality": "cef-1-normal",
                "indexed": 1,
                "datasources": [{...}]
            },
            "metadata": {
                "sources": [{...},{...}]
            },
            "usage": {
                "input_tokens": 1272,
                "output_tokens": 281,
                "total_tokens": 1553
            }
        }
        ```

    """
    name: str = "igpt_ask"
    description: str = (
        "Generate a response based on the input and related context. "
        "Use the Ask tool when you need to ask a question and get an answer grounded in the user’s complete personal information archive, "
        "including emails, messages, file attachments, and related metadata. "
        "The Ask tool is ideal for finding relevant content — "
        "such as specific documents, conversation history, key decisions, dates, or background details "
        "the user previously received or discussed."
    )
    args_schema: Type[BaseModel] = IgptAskInput
    """The schema that is passed to the model when performing tool calling."""

    params: Optional[Dict[str, Any]] = None

    api_wrapper: IgptAskAPIWrapper = Field(default_factory=IgptAskAPIWrapper)

    def __init__(self, user: Optional[str] = None, api_key: Optional[str] = None, base_url: Optional[str] = None, max_retries: int = 3, backoff_factor: int = 2, backoff_base: int = 100, **kwargs: Any) -> None:
        data = {}
        if api_key or base_url or user:
            wrapper_kwargs = {}
            if api_key:
                wrapper_kwargs["api_key"] = api_key
            if base_url:
                wrapper_kwargs["base_url"] = base_url.rstrip("/")
            if user:
                wrapper_kwargs["user"] = user
                
            wrapper_kwargs["max_retries"] = max(1, max_retries) if isinstance(max_retries, int) else 3
            wrapper_kwargs["backoff_factor"] = max(2, backoff_factor) if isinstance(backoff_factor, int) else 2

            backoff_base_ms = max(100, backoff_base) if isinstance(backoff_base, int) else 100
            wrapper_kwargs["backoff_base"] = backoff_base_ms / 1000.0 # 0.1 seconds
            data["api_wrapper"] = IgptAskAPIWrapper(**wrapper_kwargs)

        data["params"] = dict(kwargs or {})

        super().__init__(**data)

    def _run(self, user: Optional[str] = None, **kwargs: Any) -> Union[Dict[str, Any], Generator[dict, None, None]]:
        params ={}
        params.update(self.params)
        if kwargs:
            params.update(kwargs)
        try:
            res = self.api_wrapper.ask(user=user, **params)

            if isinstance(res, dict):
                if res.get("context") and not res["context"].get("datasources"):
                    warnings.warn(
                        "No datasources found. Please authorize and connect a new datasource at " 
                        "https://igpt.ai/hub/playground/connectors-authorize/"
                    )

            return res
        except ToolException:
            # Re-raise tool exceptions
            raise
        except Exception as e:
            return {"error": e}

    async def _arun(self, user: Optional[str] = None, **kwargs: Any) -> Union[Dict[str, Any], AsyncGenerator[dict, None]]:
        params ={}
        params.update(self.params)
        if kwargs:
            params.update(kwargs)
        try:
            res = await self.api_wrapper.ask_async(user=user, **params)

            if isinstance(res, dict):
                if res.get("context") and not res["context"].get("datasources"):
                    warnings.warn(
                        "No datasources found. Please authorize and connect a new datasource at " 
                        "https://igpt.ai/hub/playground/connectors-authorize/"
                    )
            
            return res
        except ToolException:
            # Re-raise tool exceptions
            raise
        except Exception as e:
            return {"error": e}




from langchain_igpt.igpt_ask import IgptAsk
from langchain_igpt.igpt_search import IgptSearch
from langchain_igpt.igpt_search_retriever import IgptSearchRetriever

__all__ = ["IgptAsk", "IgptSearch", "IgptSearchRetriever"]


from typing import Any, Dict, Optional, Type
from langchain_core.tools import BaseTool, ToolException
from pydantic import BaseModel, ConfigDict, Field
from langchain_igpt._igpt_search_api_wrapper import IgptSearchAPIWrapper
import warnings

class IgptSearchInput(BaseModel):
    model_config = ConfigDict(extra="allow")

    query: Optional[str] = Field(
        default=None,
        description=("The search query to execute.")
    )

    date_from: Optional[str] = Field(
        default=None,
        description="""Filter by start date. 

        Format must be YYYY-MM-DD (e.g., "2024-01-15" for January 15, 2024).

        Examples:
        - "2025-01-02" - Results from January 2, 2025 onwards
        - "2024-12-21" - Results from December 21, 2024 onwards

        Default is None.
        """,
    )

    date_to: Optional[str] = Field(
        default=None,
        description="""Filter by end date.

        Format must be YYYY-MM-DD (e.g., "2024-03-31" for March 31, 2024).

        Examples:
        - "2024-03-31" - Results up to and including March 31, 2024
        - "2023-12-31" - Results up to and including December 31, 2023

        Default is None.
        """,
    )

class IgptSearch(BaseTool):
    """

    Setup:
        Install `langchain-igpt` and set environment variables `IGPT_API_KEY`, `IGPT_API_USER`.

        ```bash
        pip install -U langchain-igpt
        export IGPT_API_KEY="your-api-key"
        export IGPT_API_USER="your-api-user"
        ```

    Instantiate:

        ```python
        from langchain_igpt import IgptSearch

        tool = IgptSearch(max_results=5)
        ```

    Invoke directly with args:

        ```python
        tool.invoke({"query": "invoices", "date_from":"2024-03-31", "date_to":"2025-01-02"})
        ```

        ```json
       {
            "id": "...",
            "context": {
                "indexed": 1,
                "datasources": [{...}]
            },
            "results": [{
                "id": "...",
                "timestamp": "2024-05-13T11:48:34Z",
                "type": "message",
                "from": "...",
                "subject": "...",
                "content": "..."
                },{...}],
            "durations": {
                "retrieve": 47,
                "resolve": 10
            }
        }
        ```

    """
    name: str = "igpt_search"
    description: str = (
        "Search in connected datasources. "
        "The Search tool provides direct, retrieval-only access to indexed context across your connected datasources, "
        "including emails, messages, file attachments, and related metadata. "
        "It returns raw, context-aware results without applying any reasoning, summarization, "
        "or generation - so you can inspect and use the underlying evidence as-is. "
        "Each result includes stable, structured metadata, making it well-suited for building citations, assembling lists, "
        "and implementing custom filtering or downstream processing pipelines."
    )
    args_schema: Type[BaseModel] = IgptSearchInput
    """The schema that is passed to the model when performing tool calling."""

    params: Optional[Dict[str, Any]] = None

    api_wrapper: IgptSearchAPIWrapper = Field(default_factory=IgptSearchAPIWrapper)

    def __init__(self, user: Optional[str] = None, api_key: Optional[str] = None, base_url: Optional[str] = None, max_retries: int = 3, backoff_factor: int = 2, backoff_base: int = 100, **kwargs: Any) -> None:
        data = {}
        if api_key or base_url or user:
            wrapper_kwargs = {}
            if api_key:
                wrapper_kwargs["api_key"] = api_key
            if base_url:
                wrapper_kwargs["base_url"] = base_url.rstrip("/")
            if user:
                wrapper_kwargs["user"] = user
                
            wrapper_kwargs["max_retries"] = max(1, max_retries) if isinstance(max_retries, int) else 3
            wrapper_kwargs["backoff_factor"] = max(2, backoff_factor) if isinstance(backoff_factor, int) else 2

            backoff_base_ms = max(100, backoff_base) if isinstance(backoff_base, int) else 100
            wrapper_kwargs["backoff_base"] = backoff_base_ms / 1000.0 # 0.1 seconds
            data["api_wrapper"] = IgptSearchAPIWrapper(**wrapper_kwargs)

        data["params"] = dict(kwargs or {})

        super().__init__(**data)

    def _run(self, user: Optional[str] = None, **kwargs: Any) -> Dict[str, Any]:
        params ={}
        params.update(self.params)
        if kwargs:
            params.update(kwargs)

        try:
            res = self.api_wrapper.search(user=user, **params)

            if res.get("context") and not res["context"].get("datasources"):
                warnings.warn(
                    "No datasources found. Please authorize and connect a new datasource at "
                    "https://igpt.ai/hub/playground/connectors-authorize/"
                )

            return res
        except ToolException:
            # Re-raise tool exceptions
            raise
        except Exception as e:
            return {"error": e}

    async def _arun(self, user: Optional[str] = None, **kwargs: Any) -> Dict[str, Any]:
        params ={}
        params.update(self.params)
        if kwargs:
            params.update(kwargs)
        try:
            res = await self.api_wrapper.search_async(user=user, **params)
            
            if res.get("context") and not res["context"].get("datasources"):
                warnings.warn(
                    "No datasources found. Please authorize and connect a new datasource at " 
                    "https://igpt.ai/hub/playground/connectors-authorize/"
                )

            return res
        except ToolException:
            # Re-raise tool exceptions
            raise
        except Exception as e:
            return {"error": e}
